# pylint: skip-file
"""Declares :class:`BaseProvider`."""
from django.contrib.auth import authenticate
from django.contrib.auth import login


class BaseProvider:
    """The base class for all Single Sign-on (SSO) providers."""

    @classmethod
    def fromresource(cls, resource, *args, **kwargs):
        """Create a new :class:`BaseProvider` from a provider
        resource.
        """
        raise NotImplementedError("Subclasses must override this method.")

    def __init__(self, spec):
        self._spec = spec

    def authenticate(self, request, *args, **kwargs):
        """Authenticate a :term:`Subject` using the Django authentication system."""
        return authenticate(request, *args, **kwargs) # pragma: no cover

    def login(self, request, subject):
        """Login a :term:`Subject` using the Django authentication system."""
        login(request, subject) # pragma: no cover

    def is_default(self):
        """Return a boolean indicating if the provider has marked itself as
        a default provider.
        """
        return bool(self._spec.get('is_default'))

    def is_enabled(self):
        """Return a boolean indicating if the provider is enabled."""
        return True



"""Declares :class:`GenericOAuth2Provider`."""
from .base import BaseOAuth2Provider


class GenericOAuth2Provider(BaseOAuth2Provider):
    """Generic OAuth2 provider implementation."""

    def get_asserted_subject_id(self, cs): # pylint: disable=invalid-name
        """Return the asserted subject identifier from the OAuth2 provider
        claims set.
        """
        raise NotImplementedError("Subclasses must override this method.")



"""Declares :class:`AuthenticationCtrl`."""
import functools
import logging
import time
import uuid

import ioc
from django.http import JsonResponse
from django.http import HttpResponse
from django.http import HttpResponseRedirect
from django.middleware.csrf import rotate_token
from django.urls import include
from django.urls import reverse
from django.urls import path
from django.utils.http import  url_has_allowed_host_and_scheme

from iam.ext.idc.lib import IdentityProvidersRegistry


class AuthenticationCtrl:
    """A controller that manages the authentication flow using the SAML2
    or OAuth2 protocols.
    """
    logger = logging.getLogger('iam.audit')
    redirect_cookie_name = 'idc.post_login_redirect'
    success_url_allowed_hosts = set()

    @property
    def urlpatterns(self):
        """Return the list of URL patterns exposed by :class:`AuthenticationCtrl`."""
        return [
            path('login', self.login_using_default,
                name='login.default'),
            path('error', self.nologin,
                name='nologin'),
            path('providers/default', self.get_default_provider,
                name='providers.default'),
            path('providers', self.list_providers, name='providers'),
            path('<str:using>/authorize', self.inject_provider(self.authorize),
                name='authorize'),
            path('<str:using>/login', self.inject_provider(self.login),
                name='login'),
        ]

    @classmethod
    def as_namespace(cls, base_path, namespace='iam', registry=None):
        """Instantiate a :class:`AuthenticationCtrl` instance and return an
        iterable suitable for use with :func:`django.urls.include`.

        Args:
            base_path (str): the base path to mount the URLs to. Must not
                include a leading slash.
            namespace (str): the URL namespace. Defaults to ``iam``.
            registry: a :class:`IdentityProvidersRegistry` instance. May
                be ``None``, in which case the default registry is used.

        Returns:
            iterable
        """
        ctrl = cls(registry or IdentityProvidersRegistry(
            ioc.require('iam.IdentityProviderLoaders')), namespace=namespace)
        return path(base_path, include((ctrl.urlpatterns, 'idc'), namespace=namespace))

    def __init__(self, registry, namespace):
        self.registry = registry
        self.namespace = namespace

    def authorize(self, request, provider):
        """Invoked when the user is redirected by the OAuth2 SSO provider after the
        authentication dialog.
        """
        try:
            subject = provider.get_asserted_subject(request)
        except Exception as e: # pylint: disable=W0703,C0103
            self.logger.error("Invalid response from provider '%s': %s",
                provider.name, e.args[0])
            return HttpResponse(status=500)
        if subject is not None:
            provider.login(request, subject)
            redirect_to = self.get_login_redirect(request)
        else:
            redirect_to = reverse('iam:nologin')
        response = HttpResponseRedirect(redirect_to)
        response.delete_cookie(self.redirect_cookie_name)
        return response

    def inject_provider(self, func):
        """Decorate `func` so that the provider is injected based on the
        `using` URL parameter.
        """
        @functools.wraps(func)
        def f(request, using, *args, **kwargs): # pylint: disable=C0103
            try:
                return func(request, self.registry.get(using), *args, **kwargs)
            except KeyError:
                list_url = self.get_enabled_providers_uri(request)
                dto = {
                    'id': str(uuid.uuid4()),
                    'timestamp': int(time.time() * 1000),
                    'code': "IDENTITY_PROVIDER_DOES_NOT_EXIST",
                    'message': f"The selected identity provider '{using}' does not exist.",
                    'hint': f"Inspect {list_url} for a list of enabled identity providers.",
                    'audit': {
                        'message': "This error is recorded along with your IP address."
                    }
                }
                self.logger.info(
                    "Unknown provider '%s' requested from host %s (id: %s)",
                    using, 'n.o.o.p', dto['id'])
                return JsonResponse(dto, status=404,
                    json_dumps_params={'indent': 2})

        return f

    def get_authorize_uri(self, request, using): # pylint: disable=unused-argument
        """Return the absolute URL to which the :term:`Asserting Party` must
        redirect after the login dialog.
        """
        return request.build_absolute_uri(
            reverse(f'{self.namespace}:authorize', kwargs={'using': using}))

    def get_default_provider(self, request):
        """Return a JSON response containing the default OAuth2 provider."""
        provider = self.registry.get_default_provider()
        return JsonResponse({
            'description': provider.description,
            'name': provider.name,
            'loginUrl': self.get_login_uri(request, provider.name)
            }, json_dumps_params={'indent': 2})

    def get_enabled_providers_uri(self, request):
        """Return URL yielding the enabled providers."""
        return request.build_absolute_uri(reverse(f'{self.namespace}:providers'))

    def get_login_uri(self, request, using):
        """Return the login URL for the given provider."""
        return request.build_absolute_uri(
            reverse(f'{self.namespace}:login', kwargs={'using': using}))

    def get_login_redirect(self, request):
        """Inspect the cookies on the request object and return the URL to
        which the client must redirect after a succesful authentication. Also
        inspect the ``?next`` parameter, since there may be cookies left over
        from a failed authentication attempt.
        """
        try:
            redirect_to = request.get_signed_cookie(self.redirect_cookie_name)\
                if 'next' not in request.GET\
                else request.GET['next']
        except LookupError:
            redirect_to = '/'
        url_is_safe = url_has_allowed_host_and_scheme(
            url=redirect_to,
            allowed_hosts=self.get_success_url_allowed_hosts(request),
            require_https=request.is_secure(),
        )
        return redirect_to if url_is_safe else '/'

    def get_success_url_allowed_hosts(self, request):
        """Return the set of allowed hosts to redirect to."""
        return {request.get_host(), *self.success_url_allowed_hosts}

    def list_providers(self, request):
        """Return a JSON response containing the enabled OAuth2 providers."""
        rotate_token(request)
        return JsonResponse({
            'providers': [{
            'description': x.description,
            'loginUrl': self.get_login_uri(request, x.name)
        } for x in self.registry.enabled()]}, json_dumps_params={'indent': 2})

    def login(self, request, provider):
        """Redirect the client to the identity provider login screen."""
        response = provider.start_authentication(request,
            self.get_authorize_uri(request, provider.name))
        redirect_to = self.get_login_redirect(request) or '/'
        response.set_signed_cookie(self.redirect_cookie_name, redirect_to,
            max_age=60, httponly=True)
        return response

    def login_using_default(self, request):
        """Login using the default provider."""
        return self.login(request, self.registry.get_default_provider())

    def nologin(self, request): # pylint: disable=W0613,R0201
        """Display an error message when we refuse to authenticate the user."""
        return HttpResponse("Unable to authenticate.", status=403)



"""Declares :class:`GoogleOIDCProvider`."""
import ioc

from .oidc import OIDCGenericProvider


class GoogleOIDCProvider(OIDCGenericProvider):
    """Implements OAuth2 with Google using OpenID Connect (OIDC)."""
    metadata_url = "https://accounts.google.com/.well-known/openid-configuration"
    default_scope = ["openid", "email", "profile"]

    @classmethod
    def fromenv(cls, env):
        """Instantiate a new :class:`GoogleOIDCProvider` from environment
        variables.
        """
        oauth = ioc.require('OAuth2Registry')
        client = oauth.register(
            name='google-oidc',
            client_id=env['GOOGLE_OAUTH_CLIENT_ID'],
            client_secret=env['GOOGLE_OAUTH_CLIENT_SECRET'],
            client_kwargs={
                'scope': str.split(' ', env['GOOGLE_OAUTH_SCOPE'])\
                    if 'GOOGLE_OAUTH_SCOPE' in env\
                    else cls.default_scope
            },
            server_metadata_url=cls.metadata_url
        )
        return cls(client, {'description': "Google"})

    @staticmethod
    def get_client_kwargs(spec): # pylint: disable=arguments-differ
        """Return the client keyword arguments for the specified
        provider `spec`.
        """
        kwargs = {}
        if spec.get('scope'):
            kwargs['scope'] = str.join(' ', spec['scope'])
        return kwargs

    def get_asserted_subject_id(self, cs):
        """Return the asserted subject identifier from the OAuth2 provider
        claims set.
        """
        return f"{cs['sub']}@{cs['aud']}"



# pylint: disable=C0103
"""Declares :class:`SecretsDirectoryLoader`."""
import os

import yaml

from iam.canon.idc import ResourceSchema
from ..oauth2 import get_oauth2_provider
from .base import BaseLoader


class SecretsDirectoryLoader(BaseLoader):
    """Loads identity provider configurations from the application secrets
    directory.
    """
    schema_class = ResourceSchema
    provider_factories = {
        'OAuth2Provider': get_oauth2_provider,
    }

    @property
    def secdir(self):
        """Returns the directory name holding the secrets (i.e. provider
        configurations) used by the :class:`BaseIdentityProviverRegistry`.
        """
        return os.path.join(self.__secdir, "iam.unimatrixone.io/providers")

    def __init__(self, secdir=None):
        self.__secdir = secdir or os.environ['APP_SECDIR']
        self.__schema = self.schema_class()

    def get_provider(self, resource):
        """Instantiate a new :class:`Provider` from the parameters
        specified in `resource`.
        """
        f = self.get_provider_factory(resource) # pylint: disable=C0103
        return f(resource)

    def get_provider_factory(self, resource):
        """Returns the provider implementation class for the given
        resource.
        """
        return self.provider_factories[ resource['kind'] ]

    def is_supported_resource(self, resource):
        """Return a boolean indicating if the resource is supported by this
        provider.
        """
        return resource.get('kind') in list(self.provider_factories.keys())

    def load(self):
        """Iterate over all YAML files in the secrets directory and instantiate
        providers.
        """
        providers = []
        if not os.path.exists(self.secdir): # pragma: no cover
            return providers

        for fn in os.listdir(self.secdir):
            fp = os.path.join(self.secdir, fn)
            for doc in yaml.load_all(open(fp), Loader=yaml.SafeLoader): # pylint: disable=C0103
                if not self.is_supported_resource(doc): # pragma: no cover
                    # Silently ignore unrecognized documents.
                    continue
                providers.append(self.get_provider(doc))

        return providers



"""Declares :class:`OIDCGenericProvider`."""
from .base import BaseOAuth2Provider


class OIDCGenericProvider(BaseOAuth2Provider):
    """Generic implementation of OAuth2 using the OpenID Connect (OIDC)
    protocol.
    """

    def get_asserted_subject_id(self, cs): # pylint: disable=invalid-name
        """Return the asserted subject identifier from the OAuth2 provider
        claims set.
        """
        raise NotImplementedError("Subclasses must override this method.")



"""Declares :class:`OAuth2ProviderSpecSchema`."""
import marshmallow
import marshmallow.fields

from .oauth2credentials import OAuth2CredentialsSchema
from .oauth2providermetadata import OAuth2ProviderMetadataSchema


class OAuth2ProviderSpecSchema(marshmallow.Schema):
    """Declares a schema for ``OAuth2ProviderSpec`` resources."""
    description = marshmallow.fields.String(
        required=True,
        data_key='description'
    )

    scope = marshmallow.fields.List(
        marshmallow.fields.String,
        required=False,
        default=list,
        missing=list
    )

    credentials = marshmallow.fields.Nested(
        OAuth2CredentialsSchema,
        required=True
    )

    superusers = marshmallow.fields.List(
        marshmallow.fields.String,
        required=False,
        missing=list,
        default=list
    )

    external = marshmallow.fields.Boolean(
        required=False,
        missing=False,
        default=False
    )

    is_default = marshmallow.fields.Boolean(
        required=False,
        missing=False,
        default=False,
        data_key='defaultProvider'
    )

    url = marshmallow.fields.URL(
        required=False,
        missing=None,
        default=None,
        data_key='metadataUrl'
    )

    metadata = marshmallow.fields.Nested(
        OAuth2ProviderMetadataSchema,
        required=False,
        missing=None,
        default=None,
        data_key='providerMetadata'
    )



# pylint: skip-file
from iam.ext.idc.lib.oauth2.providers.google import GoogleOIDCProvider
from .base import BaseProviderTestcase


PROVIDER_RESOURCE = """
apiVersion: testing
kind: OAuth2Provider
metadata:
  name: test-oidc-google
type: accounts.google.com
spec:
  description: Test Provider
  credentials:
    id: foo
    secret: bar
  scope:
  - email
  - profile
"""


class OIDCGenericProviderTestCase(BaseProviderTestcase):
    environ = {
        'GOOGLE_OAUTH_CLIENT_ID': "foo",
        'GOOGLE_OAUTH_CLIENT_SECRET': "bar"
    }
    provider_resource = PROVIDER_RESOURCE
    provider_class = GoogleOIDCProvider
    __test__ = True

    def test_get_asserted_subject_id(self):
        p = self.registry.get(self.resource['metadata']['name'])
        asid = p.get_asserted_subject_id({'sub': '1', 'aud': 'foo'})
        self.assertEqual(asid, "1@foo")

    def test_fromenv(self):
        p = self.provider_class.fromenv(self.environ)
        self.assertEqual(p.description, "Google")



# pylint: skip-file
from iam.ext.idc.lib.oauth2.providers.generic import GenericOAuth2Provider
from .base import BaseProviderTestcase


PROVIDER_RESOURCE = """
apiVersion: v1alpha1
kind: OAuth2Provider
metadata:
  name: test-oauth-basic
spec:
  description: Basic OAuth2
  credentials:
    id: foo
    secret: bar
  defaultProvider: true
  providerMetadata:
    authorizeUrl: https://api.twitter.com/oauth/authenticate
    baseUrl: https://api.twitter.com/1.1/
    requestToken:
      url: https://api.twitter.com/oauth/request_token
      #params: null
    accessToken:
      url: https://api.twitter.com/oauth/access_token
      #params: null
    #refreshToken:
    #  url: null
"""


class GenericOAuth2ProviderTestCase(BaseProviderTestcase):
    provider_resource = PROVIDER_RESOURCE
    provider_class = GenericOAuth2Provider
    __test__ = True



"""Declares :class:`OAuth2CredentialsSchema`."""
import marshmallow
import marshmallow.fields


class OAuth2CredentialsSchema(marshmallow.Schema):
    """Declares a schema for ``OAuth2Credentials`` resources."""
    id = marshmallow.fields.String(required=True)
    secret = marshmallow.fields.String(required=True)



# pylint: disable=R0801
"""Declares :class:`OAuth2ProviderSchema`."""
import marshmallow
import marshmallow.fields

from .oauth2providerspec import OAuth2ProviderSpecSchema
from .objectmeta import ObjectMetaSchema


class OAuth2ProviderSchema(marshmallow.Schema):
    """Declares a schema for ``OAuth2Provider`` resources."""
    apiVersion = marshmallow.fields.String(
        required=True
    )

    kind = marshmallow.fields.String(
        required=True
    )

    metadata = marshmallow.fields.Nested(
        ObjectMetaSchema,
        required=True
    )

    spec = marshmallow.fields.Nested(
        OAuth2ProviderSpecSchema,
        required=True
    )

    type = marshmallow.fields.String(
        required=False,
        missing="iam.unimatrixone.io/oidc",
        default="iam.unimatrixone.io/oidc"
    )



# pylint: skip-file
from iam.ext.idc.lib.oauth2.providers.generic import GenericOAuth2Provider
from .base import BaseProviderTestcase


PROVIDER_RESOURCE = """
apiVersion: testing
kind: OAuth2Provider
metadata:
  name: test-oidc-generic
spec:
  description: Test Provider
  credentials:
    id: foo
    secret: bar
  scope:
  - email
  - profile
  metadataUrl: https://example.com/metadata
"""


class OIDCGenericProviderTestCase(BaseProviderTestcase):
    provider_resource = PROVIDER_RESOURCE
    provider_class = GenericOAuth2Provider
    __test__ = True



"""Declares :class:`IdentityConsumerConfig`."""
import ioc
import ioc.loader
from authlib.integrations.django_client import OAuth
from django.apps import AppConfig
from django.apps import apps
from django.conf import settings

from .infra import SubjectRepository


class IdentityConsumerConfig(AppConfig):
    """Configures the :mod:`iam.ext.idc` package."""
    name = 'iam.ext.idc'
    label = 'idc'
    verbose_name = "IAM - Identity Consumer"
    default_loaders = [
        'iam.ext.idc.lib.EnvironmentLoader',
        'iam.ext.idc.lib.SecretsDirectoryLoader'
    ]

    def ready(self):
        """Invoked when the Django app registry has loaded all
        apps.
        """
        ioc.provide('OAuth2Registry', OAuth())
        ioc.provide('iam.SubjectRepository', SubjectRepository())
        ioc.provide('iam.SubjectFactory', apps.get_model('idc.AssertedSubject'))

        loaders = []
        ioc.provide('iam.IdentityProviderLoaders', loaders)
        for qualname in getattr(settings, 'IAM_IDP_LOADERS', self.default_loaders):
            cls = ioc.loader.import_symbol(qualname)
            loaders.append(cls())



# pylint: disable=no-self-use,unused-argument,invalid-name
"""Declares :class:`BaseIdentityProviverRegistry`."""
import abc
import itertools
from collections import OrderedDict


NOT_LOADED = object()


class BaseIdentityProviverRegistry(metaclass=abc.ABCMeta):
    """The base class for all identity provider registries."""

    def __init__(self, loaders, *args, **kwargs):
        self.__providers = NOT_LOADED
        self.__default = None
        self.__loaders = loaders

    def enabled(self):
        """Return a list containing all enabled providers."""
        if self.__providers == NOT_LOADED: # pragma: no cover
            self.load_providers()
        return list([x for x in self.__providers.values() if x.is_enabled()])

    def lookup(self, name):
        """Return the identity provider identified by `name`."""
        return self.__providers[name]

    def load_providers(self):
        """Load the providers of the specified :attr:`resource_kind` from
        the secrets directory.
        """
        providers = OrderedDict()
        for provider in itertools.chain(*map(lambda x: x.load(), self.__loaders)):
            if provider.is_default():
                self.__default = provider
            providers[provider.name] = provider

        self.__providers = providers

        # If there is no default provider at this point, select the first loaded
        # from the persistent configuration.
        if self.__default is None and bool(providers):
            self.__default = list(providers.values())[0]

    def get(self, name):
        """Return the identity provider identified by `name`. Load the providers
        from the persistent storage backend if not present.
        """
        if self.__providers == NOT_LOADED:
            self.load_providers()
        return self.lookup(name)

    def get_default_provider(self):
        """Return the default provider configured for this registry."""
        if self.__providers == NOT_LOADED: # pragma: no cover
            self.load_providers()
        return self.__default



"""Declares :class:`OAuth2ProviderMetadataSchema`."""
import marshmallow
import marshmallow.fields


class OAuth2ProviderMetadataSchema(marshmallow.Schema):
    """Decalres a schema for ``OAuth2ProviderMetadataS`` resources."""
    authorize_url = marshmallow.fields.URL(
        required=True,
        data_key='authorizeUrl'
    )

    base_url = marshmallow.fields.URL(
        required=True,
        data_key='baseUrl'
    )

    request_token = marshmallow.fields.Nested(
        'OAuth2TokenURLConfigSchema',
        required=True,
        data_key='requestToken'
    )

    access_token = marshmallow.fields.Nested(
        'OAuth2TokenURLConfigSchema',
        required=True,
        data_key='accessToken'
    )


class OAuth2TokenURLConfigSchema(marshmallow.Schema):
    """Declares a schema to validate metadata URLs."""
    url = marshmallow.fields.URL(required=True)
    params = marshmallow.fields.Dict(
        allow_none=True,
        required=False,
        missing=None,
        default=None
    )
